@@ -13,7 +13,6 @@ from django.db.models import QuerySet, Model
 from rest_framework.authentication import BaseAuthentication
 from rest_framework.parsers import FormParser, MultiPartParser
 from rest_framework.permissions import BasePermission
-from rest_framework.renderers import BrowsableAPIRenderer
 from rest_framework.request import Request
 from rest_framework.response import Response
 from rest_framework.viewsets import ViewSet
@@ -92,6 +91,7 @@ class JsonApiModelViewBuilder:
         self._after_delete_callback = None
         self._before_list_callback = None
         self._after_list_callback = None
+        self._before_raw_response = None
         self._is_admin = False
         if queryset is None:
             self._queryset = self._model.objects
@@ -192,7 +192,7 @@ class JsonApiModelViewBuilder:
         for rel in related:
             rel.api_version = rel.api_version.replace('.', '').replace('-', '')
             api_fixed_related.append(rel)
-            
+
         self._generic_relations[limit_to_on_retrieve].append(
             GenericRelation(field=field, related=api_fixed_related, many=many, required=required))
         return self
@@ -262,6 +262,10 @@ class JsonApiModelViewBuilder:
         self.__warn_if_method_not_available(json_api_spec_http_methods.HTTP_GET)
         return self
 
+    def before_response(self, before_raw_response: Callable[[str], str] = None) -> 'JsonApiModelViewBuilder':
+        self._before_raw_response = before_raw_response
+        return self
+
     def _get_history_urls(self) -> Sequence[partial]:
         history_builder = deepcopy(self)
         if plugins.AUTO_ADMIN_VIEWS not in history_builder._skip_plugins:
@@ -281,11 +285,11 @@ class JsonApiModelViewBuilder:
             .add_filter(name='history_id', lookups=(filter_lookups.EXACT, filter_lookups.IN)) \
             .add_filter(name='history_change_reason', lookups=(filter_lookups.EXACT, filter_lookups.IN)) \
             .add_filter(name='history_type', lookups=(filter_lookups.EXACT, filter_lookups.IN)) \
-            .get_urls(urls_prefix='history/', url_resource_name=self._resource_name)
+            .get_urls(urls_prefix='history/', url_resource_name=self._resource_name, ignore_swagger=True)
 
         return history_urls
 
-    def _get_admin_urls(self) -> Sequence[partial]:
+    def _get_admin_urls(self, ignore_swagger: bool = False) -> Sequence[partial]:
         admin_builder = deepcopy(self)
         admin_builder._skip_plugins = [plugins.AUTO_ADMIN_VIEWS]
         admin_builder._spice_queryset = None
@@ -303,11 +307,12 @@ class JsonApiModelViewBuilder:
             admin_builder._permission_classes = [*admin_builder._permission_classes, admin_permission_class]
 
         admin_urls = admin_builder._build(url_resource_name=self._resource_name, urls_prefix='admin/',
-                                          ignore_serializer=False)
+                                          ignore_serializer=False, ignore_swagger=ignore_swagger)
 
         return admin_urls
 
-    def _build(self, url_resource_name: str = '', urls_prefix: str = '', ignore_serializer: bool = False) -> Sequence[
+    def _build(self, url_resource_name: str = '', urls_prefix: str = '', ignore_serializer: bool = False,
+               ignore_swagger: bool = False) -> Sequence[
         partial]:
         method_to_serializer = {}
         if not ignore_serializer:
@@ -392,11 +397,22 @@ class JsonApiModelViewBuilder:
                 response = Response(serializer.data)
 
             if self._after_list_callback is not None:
-                response.data = self._after_list_callback(response.data)
+                response.data = self._after_list_callback(request, response.data)
             return response
 
+        class Renderer(JSONRenderer):
+            def render(inner_self, data, accepted_media_type=None, renderer_context=None):
+                response = super(Renderer, inner_self).render(
+                    data, accepted_media_type, renderer_context
+                )
+                if self._before_raw_response:
+                    response = self._before_raw_response(response)
+                if not isinstance(response, (bytes, bytearray)):
+                    return str.encode(response)
+                return response
+
         base_model_view_set = type(f'{self._resource_name}JSONApiModelViewSet{self._api_version}', (ModelViewSet,), {
-            'renderer_classes': (JSONRenderer, BrowsableAPIRenderer),
+            'renderer_classes': (Renderer,),
             'parser_classes': (JSONParser, FormParser, MultiPartParser),
             'metadata_class': JSONAPIMetadata,
             'pagination_class': LimitedJsonApiPageNumberPagination,
@@ -405,6 +421,9 @@ class JsonApiModelViewBuilder:
             'resource_name': self._resource_name,
         })
 
+        if ignore_swagger:
+            base_model_view_set.swagger_schema = None
+
         def get_queryset(view):
             if self._queryset is None:
                 queryset = super(view.__class__, view).get_queryset()
@@ -426,6 +445,9 @@ class JsonApiModelViewBuilder:
                                          'lookup_field': pk_name
                                      })
 
+            if ignore_swagger:
+                relationship_view.swagger_schema = None
+
             list_method_view_set = type(f'List{self._resource_name}ViewSet{self._api_version}', (base_model_view_set,),
                                         {
                                             'get_queryset': get_queryset,
@@ -500,12 +522,13 @@ class JsonApiModelViewBuilder:
                 pass
 
         if plugins.AUTO_ADMIN_VIEWS not in self._skip_plugins:
-            urls.extend(self._get_admin_urls())
+            urls.extend(self._get_admin_urls(ignore_swagger=ignore_swagger))
 
         return urls
 
-    def get_urls(self, url_resource_name: str = '', urls_prefix: str = '') -> Sequence[partial]:
-        return self._build(url_resource_name=url_resource_name, urls_prefix=urls_prefix)
+    def get_urls(self, url_resource_name: str = '', urls_prefix: str = '', ignore_swagger: bool = False) -> Sequence[
+        partial]:
+        return self._build(url_resource_name=url_resource_name, urls_prefix=urls_prefix, ignore_swagger=ignore_swagger)
 
 
 class JsonApiResourceViewBuilder:
@@ -517,6 +540,7 @@ class JsonApiResourceViewBuilder:
                  permission_classes: Sequence[Type[BasePermission]] = None,
                  authentication_classes: Sequence[Type[BaseAuthentication]] = None,
                  raw_items=False,
+                 only_callbacks: Optional[bool] = False,
                  page_size: int = 50):
         self._allowed_methods = [*allowed_methods]
         self._resource_name = action_name
@@ -531,7 +555,9 @@ class JsonApiResourceViewBuilder:
         self._on_delete_callback = None
         self._on_list_callback = None
         self._on_get_callback = None
+        self._before_raw_response = None
         self._page_size = page_size
+        self._only_callbacks = only_callbacks
 
     def __warn_if_method_not_available(self, method: str):
         if method not in self._allowed_methods:
@@ -566,7 +592,12 @@ class JsonApiResourceViewBuilder:
         self.__warn_if_method_not_available(json_api_spec_http_methods.HTTP_GET)
         return self
 
-    def _build(self, url_resource_name: str = '', urls_prefix: str = '', urls_suffix: str = '') -> Sequence[partial]:
+    def before_response(self, before_raw_response: Callable[[str], str] = None) -> 'JsonApiResourceViewBuilder':
+        self._before_raw_response = before_raw_response
+        return self
+
+    def _build(self, url_resource_name: str = '', urls_prefix: str = '', urls_suffix: str = '',
+               ignore_swagger: bool = False) -> Sequence[partial]:
         def destroy(view, request, *args, **kwargs):
             identifier = kwargs.get(self._unique_identifier, None)
 
@@ -644,36 +675,57 @@ class JsonApiResourceViewBuilder:
                                     status=status)
                 return Response(data=data, status=status)
 
-        patch_view_set = type(f'{self._resource_name}ChangeJSONApiActionViewSet{self._api_version}', (ViewSet,), {
-            'renderer_classes': (JSONRenderer, BrowsableAPIRenderer),
-            'parser_classes': (JSONParser, FormParser, MultiPartParser),
-            'metadata_class': JSONAPIMetadata,
-            'pagination_class': LimitedJsonApiPageNumberPagination,
-            'filter_backends': (
-                QueryParameterValidationFilter, OrderingFilter, JsonApiSearchFilter),
-            'resource_name': self._resource_name,
-            'http_method_names': list(map(lambda method: method.lower(), self._allowed_methods)) + ['head', 'options'],
-            'permission_classes': self._permission_classes,
-            'authentication_classes': self._authentication_classes,
-            'update': update if self._on_update_callback else None,
-            'destroy': destroy if self._on_delete_callback else None,
-            'get': get if self._on_get_callback else None
-        })
+        class Renderer(JSONRenderer):
+            def render(inner_self, data, accepted_media_type=None, renderer_context=None):
+                response = super(Renderer, inner_self).render(
+                    data, accepted_media_type, renderer_context
+                )
+                if self._before_raw_response:
+                    response = self._before_raw_response(response)
+                if not isinstance(response, (bytes, bytearray)):
+                    return str.encode(response)
+                return response
 
-        get_view_set = type(f'{self._resource_name}RetrieveJSONApiActionViewSet{self._api_version}', (ViewSet,), {
-            'renderer_classes': (JSONRenderer, BrowsableAPIRenderer),
-            'parser_classes': (JSONParser, FormParser, MultiPartParser),
-            'metadata_class': JSONAPIMetadata,
-            'pagination_class': LimitedJsonApiPageNumberPagination,
-            'filter_backends': (
-                QueryParameterValidationFilter, OrderingFilter, JsonApiSearchFilter),
-            'resource_name': None,
-            'http_method_names': list(map(lambda method: method.lower(), self._allowed_methods)) + ['head', 'options'],
-            'permission_classes': self._permission_classes,
-            'authentication_classes': self._authentication_classes,
-            'list': _list if self._on_list_callback else None,
-            'create': create if self._on_create_callback else None,
-        })
+        patch_view_set = None
+        if any([self._on_update_callback, self._on_delete_callback, self._on_get_callback]) or not self._only_callbacks:
+            patch_view_set = type(f'{self._resource_name}ChangeJSONApiActionViewSet{self._api_version}', (ViewSet,), {
+                'renderer_classes': (Renderer,),
+                'parser_classes': (JSONParser, FormParser, MultiPartParser),
+                'metadata_class': JSONAPIMetadata,
+                'pagination_class': LimitedJsonApiPageNumberPagination,
+                'filter_backends': (
+                    QueryParameterValidationFilter, OrderingFilter, JsonApiSearchFilter),
+                'resource_name': self._resource_name,
+                'http_method_names': list(map(lambda method: method.lower(), self._allowed_methods)) + ['head',
+                                                                                                        'options'],
+                'permission_classes': self._permission_classes,
+                'authentication_classes': self._authentication_classes,
+                'update': update if self._on_update_callback else None,
+                'destroy': destroy if self._on_delete_callback else None,
+                'get': get if self._on_get_callback else None
+            })
+            if ignore_swagger:
+                patch_view_set.swagger_schema = None
+
+        get_view_set = None
+        if any([self._on_list_callback, self._on_create_callback]) or not self._only_callbacks:
+            get_view_set = type(f'{self._resource_name}RetrieveJSONApiActionViewSet{self._api_version}', (ViewSet,), {
+                'renderer_classes': (Renderer,),
+                'parser_classes': (JSONParser, FormParser, MultiPartParser),
+                'metadata_class': JSONAPIMetadata,
+                'pagination_class': LimitedJsonApiPageNumberPagination,
+                'filter_backends': (
+                    QueryParameterValidationFilter, OrderingFilter, JsonApiSearchFilter),
+                'resource_name': None,
+                'http_method_names': list(map(lambda method: method.lower(), self._allowed_methods)) + ['head',
+                                                                                                        'options'],
+                'permission_classes': self._permission_classes,
+                'authentication_classes': self._authentication_classes,
+                'list': _list if self._on_list_callback else None,
+                'create': create if self._on_create_callback else None,
+            })
+            if ignore_swagger:
+                get_view_set.swagger_schema = None
 
         urls = []
 
@@ -689,19 +741,25 @@ class JsonApiResourceViewBuilder:
         if len(url_resource_name) == 0:
             url_resource_name = self._resource_name
 
-        urls.extend([
-            url(rf'^{urls_prefix}{url_resource_name}{urls_suffix}$',
-                get_view_set.as_view({'get': 'list', 'post': 'create'}, name=f'list_{self._resource_name}'),
-                name=f'list-{self._resource_name}{self._api_version}'),
-            url(rf'^{urls_prefix}{url_resource_name}{urls_suffix}/(?P<{self._unique_identifier}>[^/.]+)/$',
-                patch_view_set.as_view({'get': 'get', 'patch': 'update', 'delete': 'destroy'},
-                                       name=f'get_{self._resource_name}'),
-                name=f'{self._resource_name}{self._api_version}-detail')
-        ])
+        if get_view_set is not None:
+            urls.extend([
+                url(rf'^{urls_prefix}{url_resource_name}{urls_suffix}$',
+                    get_view_set.as_view({'get': 'list', 'post': 'create'}, name=f'list_{self._resource_name}'),
+                    name=f'list-{self._resource_name}{self._api_version}')
+            ])
+        if patch_view_set is not None:
+            urls.extend([
+                url(rf'^{urls_prefix}{url_resource_name}{urls_suffix}/(?P<{self._unique_identifier}>[^/.]+)/$',
+                    patch_view_set.as_view({'get': 'get', 'patch': 'update', 'delete': 'destroy'},
+                                           name=f'get_{self._resource_name}'),
+                    name=f'{self._resource_name}{self._api_version}-detail')
+            ])
         return urls
 
-    def get_urls(self, url_resource_name: str = '', urls_prefix: str = '', urls_suffix: str = '') -> Sequence[partial]:
-        return self._build(url_resource_name=url_resource_name, urls_prefix=urls_prefix, urls_suffix=urls_suffix)
+    def get_urls(self, url_resource_name: str = '', urls_prefix: str = '', urls_suffix: str = '',
+                 ignore_swagger: bool = False) -> Sequence[partial]:
+        return self._build(url_resource_name=url_resource_name, urls_prefix=urls_prefix, urls_suffix=urls_suffix,
+                           ignore_swagger=ignore_swagger)
 
 
 def json_api_view(resource_name: str,
@@ -712,6 +770,7 @@ def json_api_view(resource_name: str,
                   urls_prefix: str = '',
                   urls_suffix: str = '',
                   multiple_resource=True,
+                  raw_items=False,
                   page_size: int = 50) -> FunctionType:
     def decorator(func: Callable[[Request], Tuple[Dict, int]]):
         builder = JsonApiResourceViewBuilder(action_name=resource_name,
@@ -719,7 +778,9 @@ def json_api_view(resource_name: str,
                                              allowed_methods=[method],
                                              permission_classes=permission_classes,
                                              authentication_classes=authentication_classes,
-                                             page_size=page_size)
+                                             raw_items=raw_items,
+                                             page_size=page_size,
+                                             only_callbacks=True)
         if method == json_api_spec_http_methods.HTTP_GET and multiple_resource:
             return builder.on_list(list_callback=func) \
                 .get_urls(urls_prefix=urls_prefix, urls_suffix=urls_suffix)
